@@ -12,11 +12,12 @@ The load balancing strategy is randomized, biased by the number of cached connec
 This inherently provides limited failover support, but applications must still handle exceptions as desired.
 """
 
-import gzip
+from future_builtins import map
+import warnings
 import random
 import itertools
 import collections
-from cStringIO import StringIO
+import io, gzip
 import httplib, urllib
 try:
     import simplejson as json
@@ -29,16 +30,20 @@ class Response(httplib.HTTPResponse):
         httplib.HTTPResponse.begin(self)
         self.body = self.read()
         self.close()
-        if self.getheader('content-encoding') == 'gzip':
-            self.body = gzip.GzipFile(fileobj=StringIO(self.body)).read()
+        self.time = float(self.getheader('x-response-time', 'nan'))
+        if 'gzip' in self.getheader('content-encoding', ''):
+            self.body = gzip.GzipFile(fileobj=io.BytesIO(self.body)).read()
     def __nonzero__(self):
         "Return whether status is successful."
         return httplib.OK <= self.status < httplib.MULTIPLE_CHOICES
     def __call__(self):
         "Return evaluated response body or raise exception."
         body = self.body
-        if body and self.getheader('content-type') == 'text/x-json':
+        if body and 'text/x-json' in self.getheader('content-type'):
             body = json.loads(body)
+        code, agent, text = self.getheader('warning', '  ').split(' ', 2)
+        if agent == 'lupyne':
+            warnings.warn(json.loads(text))
         if self:
             return body
         raise httplib.HTTPException(self.status, self.reason, body)
@@ -54,22 +59,32 @@ class Resource(httplib.HTTPConnection):
                 for name, value in body.items()))
             headers.update({'content-length': str(len(body)), 'content-type': 'application/x-www-form-urlencoded'})
         httplib.HTTPConnection.request(self, method, path, body, headers)
-    def __call__(self, method, path, body=None):
-        "Send request and return evaluated response body."
+    def call(self, method, path, body=None):
+        "Send request and return completed `response`_."
         self.request(method, path, body)
-        return self.getresponse()()
+        return self.getresponse()
+    def multicall(self, *requests):
+        "Pipeline requests (method, path[, body]) and return completed responses."
+        responses = []
+        for request in requests:
+            self.request(*request)
+            responses.append(self.response_class(self.sock, self.debuglevel, self.strict, self._method))
+            self._HTTPConnection__state = 'Idle'
+        for response in responses:
+            response.begin()
+        return responses
     def get(self, path, **params):
         if params:
             path += '?' + urllib.urlencode(params, doseq=True)
-        return self('GET', path)
+        return self.call('GET', path)()
     def post(self, path, **body):
-        return self('POST', path, body)
+        return self.call('POST', path, body)()
     def put(self, path, **body):
-        return self('PUT', path, body)
+        return self.call('PUT', path, body)()
     def delete(self, path, **params):
         if params:
             path += '?' + urllib.urlencode(params, doseq=True)
-        return self('DELETE', path)
+        return self.call('DELETE', path)()
 
 class Resources(dict):
     """Thread-safe mapping of hosts to optionally persistent resources.
@@ -82,7 +97,7 @@ class Resources(dict):
     def __init__(self, hosts, limit=0):
         self.update((host, self.Manager(maxlen=limit)) for host in hosts)
     def request(self, host, method, path, body=None):
-        "Send request to given host and return exclusive resource."
+        "Send request to given host and return exclusive `resource`_."
         try:
             resource = self[host].popleft()
         except IndexError:
@@ -90,11 +105,11 @@ class Resources(dict):
         resource.request(method, path, body)
         return resource
     def getresponse(self, host, resource):
-        "Return response and release resource."
+        "Return `response`_ and release `resource`_ if request completed."
         response = resource.getresponse()
         if response.status != httplib.REQUEST_TIMEOUT:
             self[host].append(resource)
-        return response
+            return response
     def priority(self, host):
         "Return priority for host.  None may be used to eliminate from consideration."
         return -len(self[host])
@@ -106,34 +121,31 @@ class Resources(dict):
         priorities.pop(None, None)
         return random.choice(priorities[min(priorities)])
     def unicast(self, method, path, body=None, hosts=()):
-        "Send request and return response from any host, optionally from given subset."
+        "Send request and return `response`_ from any host, optionally from given subset."
         host = self.choice(tuple(hosts) or self)
-        while True:
-            resource = self.request(host, method, path, body)
-            response = self.getresponse(host, resource)
-            if response.status != httplib.REQUEST_TIMEOUT:
-                return response
+        resource = self.request(host, method, path, body)
+        response = self.getresponse(host, resource)
+        return resource.call(method, path, body) if response is None else response
     def broadcast(self, method, path, body=None, hosts=()):
         "Send requests and return responses from all hosts, optionally from given subset."
         hosts = tuple(hosts) or self
-        responses = {}
-        resources = dict((host, self.request(host, method, path, body)) for host in hosts)
-        while resources:
-            for host in list(resources):
-                resource = resources.pop(host)
-                response = responses[host] = self.getresponse(host, resource)
-                if response.status == httplib.REQUEST_TIMEOUT:
-                    resources[host] = self.request(host, method, path, body)
-        return map(responses.__getitem__, hosts)
+        resources = [self.request(host, method, path, body) for host in hosts]
+        responses = list(map(self.getresponse, hosts, resources))
+        indices = [index for index, response in enumerate(responses) if response is None]
+        for index in indices:
+            resources[index].request(method, path, body)
+        for index in indices:
+            responses[index] = resources[index].getresponse()
+        return responses
 
 class Shards(dict):
-    """Mapping of keys to host clusters, with associated resources.
+    """Mapping of keys to host clusters, with associated `resources`_.
     
     :param items: host, key pairs
     :param limit: maximum number of cached connections per host
     :param multimap: mapping of hosts to multiple keys
     """
-    choice = Resources.choice.im_func
+    choice = Resources.__dict__['choice']
     def __init__(self, items=(), limit=0, **multimap):
         pairs = ((host, key) for host in multimap for key in multimap[host])
         for host, key in itertools.chain(items, pairs):
@@ -141,11 +153,11 @@ class Shards(dict):
         self.resources = Resources(itertools.chain(*self.values()), limit)
     def priority(self, hosts):
         "Return combined priority for hosts."
-        priorities = map(self.resources.priority, hosts)
+        priorities = list(map(self.resources.priority, hosts))
         if None not in priorities:
-            return len(hosts), sum(priorities)
+            return len(priorities), sum(priorities)
     def unicast(self, key, method, path, body=None):
-        "Send request and return response from any host for corresponding key."
+        "Send request and return `response`_ from any host for corresponding key."
         return self.resources.unicast(method, path, body, self[key])
     def broadcast(self, key, method, path, body=None):
         "Send requests and return responses from all hosts for corresponding key."
@@ -154,7 +166,7 @@ class Shards(dict):
         """Send requests and return responses from a minimal subset of hosts which cover all corresponding keys.
         Response overlap is possible depending on partitioning.
         """
-        shards = [frozenset()]
+        shards = frozenset(),
         for key in keys:
             shards = set(hosts.union([host]) for hosts, host in itertools.product(shards, self[key]))
         return self.resources.broadcast(method, path, body, self.choice(shards))
