@@ -2,6 +2,8 @@
 Wrappers for lucene Fields and Documents.
 """
 
+from future_builtins import map, zip
+import warnings
 import itertools
 import datetime
 import collections
@@ -12,7 +14,7 @@ class Field(object):
     """Saved parameters which can generate lucene Fields given values.
     
     :param name: name of field
-    :param store, index, termvector: field parameters, expressed as bools or strs, with lucene defaults
+    :param store,index,termvector: field parameters, expressed as bools or strs, with lucene defaults
     :param attrs: additional attributes to set on the field
     """
     def __init__(self, name, store=False, index='analyzed', termvector=False, **attrs):
@@ -54,7 +56,7 @@ class PrefixField(Field):
     The customizable component field names are expressed as slices.
     Original value may be stored for convenience.
     
-    :param start, stop, step: optional slice parameters of the prefix depths (not indices)
+    :param start,stop,step: optional slice parameters of the prefix depths (not indices)
     """
     def __init__(self, name, start=1, stop=None, step=1, index=True, **kwargs):
         Field.__init__(self, name, index=index, **kwargs)
@@ -78,7 +80,7 @@ class PrefixField(Field):
         if self.store != lucene.Field.Store.NO:
             for value in values:
                 yield lucene.Field(self.name, value, self.store, lucene.Field.Index.NO)
-        values = map(self.split, values)
+        values = list(map(self.split, values))
         for depth in self.indices(max(map(len, values))):
             name = self.getname(depth)
             for value in sorted(set(value[:depth] for value in values if len(value) > (depth-self.depths.step))):
@@ -118,6 +120,9 @@ class DateTimeField(PrefixField):
     """Field which indexes each datetime component in sortable ISO format: Y-m-d H:M:S.
     Works with datetimes, dates, and any object whose string form is a prefix of ISO.
     """
+    def __init__(self, *args, **kwargs):
+        warnings.warn('engine.DateTimeField will be replaced with engine.numeric.DateTimeField.', DeprecationWarning)
+        PrefixField.__init__(self, *args, **kwargs)
     def split(self, text):
         "Return immutable sequence of datetime components."
         words = (words.split(char) for words, char in zip(text.split(), '-:'))
@@ -151,10 +156,10 @@ class DateTimeField(PrefixField):
         """Return optimal union of date range queries.
         May produce invalid dates, but the query is still correct.
         """
-        dates = (map(float, self.split(str(date))) for date in (start, stop))
+        dates = (list(map(float, self.split(str(date)))) for date in (start, stop))
         items = []
         for dates in self._range(*dates):
-            items.append(tuple(self.join(map('{0:02g}'.format, date)) for date in dates))
+            items.append(tuple(self.join(tuple(map('{0:02g}'.format, date))) for date in dates))
         queries = [PrefixField.range(self, str(start), items[0][1], lower=lower)]
         queries += [PrefixField.range(self, *item) for item in items[1:-1]]
         queries.append(PrefixField.range(self, items[-1][0], str(stop), upper=upper))
@@ -163,7 +168,7 @@ class DateTimeField(PrefixField):
         """Return date range query within time span of date.
         
         :param date: origin date or tuple
-        :param days, delta: timedelta parameters
+        :param days,delta: timedelta parameters
         """
         if not isinstance(date, datetime.date):
             date = tuple(date) + (None, 1, 1)[len(date):]
@@ -174,7 +179,7 @@ class DateTimeField(PrefixField):
         """Return date range query within current time and delta.
         If the delta is an exact number of days, then dates will be used.
         
-        :param days, weeks: number of days to offset from today
+        :param days,weeks: number of days to offset from today
         :param utc: optionally use utc instead of local time
         :param delta: additional timedelta parameters
         """
@@ -187,14 +192,11 @@ class Document(object):
     """Delegated lucene Document.
     Provides mapping interface of field names to values, but duplicate field names are allowed.
     """
-    def __init__(self, doc=None):
-        self.doc = lucene.Document() if doc is None else doc
-    def add(self, name, value, cls=Field, **params):
-        for field in cls(name, **params).items(value):
-            self.doc.add(field)
+    def __init__(self, doc):
+        self.doc = doc
     def fields(self):
         "Generate lucene Fields."
-        return itertools.imap(lucene.Field.cast_, self.doc.getFields())
+        return map(lucene.Field.cast_, self.doc.getFields())
     def __len__(self):
         return self.doc.getFields().size()
     def __contains__(self, name):
@@ -215,12 +217,8 @@ class Document(object):
         "Return field value if present, else default."
         value = self.doc[name]
         return default if value is None else value
-    def __delitem__(self, name):
-        self.doc.removeFields(name)
     def getlist(self, name):
         "Return list of all values for given field."
-        if lucene.VERSION <= '2.4.1': # memory leak in string array
-            return [field.stringValue() for field in self.doc.getFields(name)]
         return list(self.doc.getValues(name))
     def dict(self, *names, **defaults):
         """Return dict representation of document.
@@ -254,20 +252,21 @@ class Hits(object):
     :param ids: ordered doc ids
     :param scores: ordered doc scores
     :param count: total number of hits
+    :param maxscore: maximum score
     :param fields: optional field selectors
     """
-    def __init__(self, searcher, ids, scores, count=0, fields=None):
+    def __init__(self, searcher, ids, scores, count=None, maxscore=None, fields=None):
         self.searcher = searcher
         self.ids, self.scores = ids, scores
-        self.count = count or len(self)
+        self.count, self.maxscore = count, maxscore
         self.fields = lucene.MapFieldSelector(fields) if isinstance(fields, collections.Iterable) else fields
     def __len__(self):
         return len(self.ids)
     def __getitem__(self, index):
         id, score = self.ids[index], self.scores[index]
         if isinstance(index, slice):
-            return type(self)(self.searcher, id, score, self.count, self.fields)
+            return type(self)(self.searcher, id, score, self.count, self.maxscore, self.fields)
         return Hit(self.searcher.doc(id, self.fields), id, score)
     def items(self):
         "Generate zipped ids and scores."
-        return itertools.izip(self.ids, self.scores)
+        return zip(self.ids, self.scores)
