@@ -4,12 +4,13 @@ Wrappers for lucene Index{Read,Search,Writ}ers.
 The final `Indexer`_ classes exposes a high-level Searcher and Writer.
 """
 
+from future_builtins import map, zip
+import os
 import itertools, operator
 import contextlib
 import abc, collections
-import warnings
 import lucene
-from .queries import Query, HitCollector, Filter, Highlighter, SpellChecker, SpellParser
+from .queries import Query, HitCollector, Filter, SortField, Highlighter, SpellChecker, SpellParser
 from .documents import Field, Document, Hits
 
 class Atomic(object):
@@ -113,19 +114,15 @@ class Analyzer(lucene.PythonAnalyzer):
         """Return parsed lucene Query.
         
         :param query: query string
-        :param field: default query field name, sequence of names, or (in lucene 2.9 or higher) boost mapping
+        :param field: default query field name, sequence of names, or boost mapping
         :param op: default query operator ('or', 'and')
         :param version: lucene Version string, leave blank for deprecated constructor
         :param parser: custom PythonQueryParser class
         :param attrs: additional attributes to set on the parser
         """
         # parsers aren't thread-safe (nor slow), so create one each time
-        try:
-            version = getattr(lucene.Version, 'LUCENE_' + version.replace('.', ''))
-        except AttributeError:
-            args = []
-        else:
-            args = [] if lucene.VERSION.startswith('2.9') and parser is not None else [version]
+        version = getattr(lucene.Version, 'LUCENE_' + version.replace('.', ''))
+        args = [] if lucene.VERSION.startswith('2.9') and parser is not None else [version]
         if isinstance(field, collections.Mapping):
             boosts = lucene.HashMap()
             for key in field:
@@ -139,7 +136,7 @@ class Analyzer(lucene.PythonAnalyzer):
         for name, value in attrs.items():
             setattr(parser, name, value)
         if isinstance(parser, lucene.MultiFieldQueryParser):
-            return parser.parse(parser, query) # bug in method binding
+            return lucene.MultiFieldQueryParser.parse(parser, query) # bug in method binding
         try:
             return parser.parse(query)
         finally:
@@ -169,6 +166,32 @@ class IndexReader(object):
     def directory(self):
         "reader's lucene Directory"
         return self.indexReader.directory()
+    def copy(self, dest, query=None, exclude=None, optimize=False):
+        """Copy the index to the destination directory.
+        Optimized to use hard links if the destination is a file system path.
+        
+        :param dest: destination directory path or lucene Directory
+        :param query: optional lucene Query to select documents
+        :param exclude: optional lucene Query to exclude documents
+        :param optimize: optionally optimize destination index
+        """
+        if isinstance(dest, lucene.Directory):
+            lucene.Directory.copy(self.directory, dest, False)
+        else:
+            src = lucene.FSDirectory.cast_(self.directory).file.path
+            os.path.isdir(dest) or os.makedirs(dest)
+            for filename in set(self.directory.listAll()).difference(['write.lock']):
+                os.link(os.path.join(src, filename), os.path.join(dest, filename))
+        with contextlib.closing(IndexWriter(dest)) as writer:
+            if query:
+                writer.delete(Query(lucene.MatchAllDocsQuery) - query)
+            if exclude:
+                writer.delete(exclude)
+            writer.commit()
+            writer.expungeDeletes()
+            if optimize:
+                writer.optimize()
+        return len(writer)
     def count(self, name, value):
         "Return number of documents with given term."
         return self.docFreq(lucene.Term(name, value))
@@ -202,6 +225,23 @@ class IndexReader(object):
                     break
                 yield (text, termenum.docFreq()) if counts else text
                 term = termenum.next() and termenum.term()
+    def numbers(self, name, step=0, type=int, counts=False):
+        """Generate decoded numeric term values, optionally with frequency counts.
+        
+        :param name: field name
+        :param step: precision step to select terms
+        :param type: int or float
+        :param counts: include frequency counts
+        """
+        term = lucene.Term(name, chr(ord(' ') + step))
+        decode = lucene.NumericUtils.prefixCodedToLong
+        convert = lucene.NumericUtils.sortableLongToDouble if issubclass(type, float) else int
+        with contextlib.closing(lucene.PrefixTermEnum(self.indexReader, term)) as termenum:
+            term = termenum.term()
+            while term:
+                value = convert(decode(term.text()))
+                yield (value, termenum.docFreq()) if counts else value
+                term = termenum.next() and termenum.term()
     def docs(self, name, value, counts=False):
         "Generate doc ids which contain given term, optionally with frequency counts."
         with contextlib.closing(self.termDocs(lucene.Term(name, value))) as termdocs:
@@ -229,16 +269,10 @@ class IndexReader(object):
         Map values into a list for speed optimization.
         
         :param name: field name
-        :param type: type of field values compatible with lucene FieldCache
-        :param parser: callable applied to field values, available in lucene 2.9 or higher
+        :param type: type object or name compatible with FieldCache
+        :param parser: lucene FieldCache.Parser or callable applied to field values
         """
-        type = getattr(type, '__name__', type).capitalize()
-        method = getattr(lucene.FieldCache.DEFAULT, 'get{0}s'.format(type))
-        if parser is None:
-            return method(self.indexReader, name)
-        bases = getattr(lucene, 'Python{0}Parser'.format(type)),
-        parser = object.__class__('', bases, {'parse' + type: staticmethod(parser)})
-        return method(self.indexReader, name, parser())
+        return SortField(name, type, parser).comparator(self.indexReader)
     def spans(self, query, positions=False, payloads=False):
         """Generate docs with occurrence counts for a span query.
         
@@ -263,18 +297,14 @@ class IndexReader(object):
                 yield doc, sum(1 for span in spans)
     def termvector(self, id, field, counts=False):
         "Generate terms for given doc id and field, optionally with frequency counts."
-        if lucene.VERSION <= '2.4.1':
-            warnings.warn('TermFreqVector.terms leaks memory')
         tfv = self.getTermFreqVector(id, field)
-        return itertools.izip(tfv.terms, tfv.termFrequencies) if counts else iter(tfv.terms)
+        return zip(tfv.terms, tfv.termFrequencies) if counts else iter(tfv.terms)
     def positionvector(self, id, field, offsets=False):
         "Generate terms and positions for given doc id and field, optionally with character offsets."
-        if lucene.VERSION <= '2.4.1':
-            warnings.warn('TermPositionVector.terms leaks memory')
         tpv = lucene.TermPositionVector.cast_(self.getTermFreqVector(id, field))
         for index, term in enumerate(tpv.terms):
             if offsets:
-                yield term, map(operator.attrgetter('startOffset', 'endOffset'), tpv.getOffsets(index))
+                yield term, list(map(operator.attrgetter('startOffset', 'endOffset'), tpv.getOffsets(index)))
             else:
                 yield term, list(tpv.getTermPositions(index))
     def morelikethis(self, doc, *fields, **attrs):
@@ -295,11 +325,10 @@ class Searcher(object):
     def __init__(self, arg, analyzer=None):
         super(Searcher, self).__init__(arg)
         if analyzer is None:
-            analyzer = lucene.StandardAnalyzer(lucene.Version.LUCENE_CURRENT) if hasattr(lucene, 'Version') else lucene.StandardAnalyzer()
+            analyzer = lucene.StandardAnalyzer(lucene.Version.LUCENE_CURRENT)
             self.shared.add(analyzer)
         self.analyzer = analyzer
-        self.filters = {}
-        self.spellcheckers = {}
+        self.filters, self.sorters, self.spellcheckers = {}, {}, {}
     def __del__(self):
         if hash(self):
             self.close()
@@ -307,16 +336,14 @@ class Searcher(object):
         self.shared = closing()
         for directory in directories:
             if isinstance(directory, basestring):
-                if hasattr(lucene.FSDirectory, 'open'):
-                    directory = lucene.FSDirectory.open(lucene.File(directory))
-                else:
-                    directory = lucene.FSDirectory.getDirectory(directory)
+                directory = lucene.FSDirectory.open(lucene.File(directory))
                 self.shared.add(directory)
             yield directory
-    def reopen(self, filters=False, spellcheckers=False):
-        """Return current Searcher, only creating a new one if necessary.
+    def reopen(self, filters=False, sorters=False, spellcheckers=False):
+        """Return current `Searcher`_, only creating a new one if necessary.
         
         :param filters: refresh cached facet :attr:`filters`
+        :param sorters: refresh cached :attr:`sorters` with associated parsers
         :param spellcheckers: refresh cached :attr:`spellcheckers`
         """
         reader = self.indexReader.reopen()
@@ -324,11 +351,18 @@ class Searcher(object):
             return self
         other = type(self)(reader, self.analyzer)
         other.shared, other.owned = self.shared, closing([reader])
+        other.filters.update((key, value if isinstance(value, lucene.Filter) else dict(value)) for key, value in self.filters.items())
         if filters:
-            other.facets([], *self.filters)
+            other.facets([], *other.filters)
+        other.sorters = dict(self.sorters)
+        if sorters:
+            for field in self.sorters:
+                other.comparator(field)
         if spellcheckers:
             for field in self.spellcheckers:
-                other.suggest(field, '')
+                other.spellchecker(field)
+        else:
+            other.spellcheckers = dict(self.spellcheckers)
         return other
     def __getitem__(self, id):
         return Document(self.doc(id))
@@ -340,7 +374,7 @@ class Searcher(object):
             return query
         if spellcheck:
             kwargs['parser'], kwargs['searcher'] = SpellParser, self
-        return Analyzer.parse.im_func(self.analyzer, query, **kwargs)
+        return Analyzer.__dict__['parse'](self.analyzer, query, **kwargs)
     def highlighter(self, query, span=True, formatter=None, encoder=None, field=None):
         "Return `Highlighter`_ specific to the searcher's analyzer and index."
         query = self.parse(query, field=field or '')
@@ -366,9 +400,9 @@ class Searcher(object):
         :param options: additional :meth:`search` options
         """
         if len(query) <= 1:
-            return self.search(*query, count=1, **options).count
+            return self.search(*query, count=1, sort=lucene.Sort.INDEXORDER, **options).count
         return self.docFreq(lucene.Term(*query))
-    def search(self, query=None, filter=None, count=None, sort=None, reverse=False, **parser):
+    def search(self, query=None, filter=None, count=None, sort=None, reverse=False, scores=False, maxscore=False, timeout=None, **parser):
         """Run query and return `Hits`_.
         
         :param query: query string or lucene Query
@@ -376,27 +410,46 @@ class Searcher(object):
         :param count: maximum number of hits to retrieve
         :param sort: if count is given, lucene Sort parameters, else a callable key
         :param reverse: reverse flag used with sort
+        :param scores: compute scores for candidate results when using lucene Sort
+        :param maxscore: compute maximum score of all results when using lucene Sort
+        :param timeout: stop search after elapsed number of seconds
         :param parser: :meth:`Analyzer.parse` options
         """
         query = lucene.MatchAllDocsQuery() if query is None else self.parse(query, **parser)
         if not isinstance(filter, (lucene.Filter, type(None))):
             filter = Filter(filter)
-        # use custom HitCollector if all results are necessary, otherwise let lucene's TopDocs handle it
+        weight = query.weight(self)
+        # use custom HitCollector if all results are necessary, otherwise use lucene's TopDocsCollectors
         if count is None:
             collector = HitCollector()
-            super(Searcher, self).search(query, filter, collector)
-            try:
-                return Hits(self, *collector.sorted(key=sort, reverse=reverse))
-            finally:
-                collector.finalize()
-        if isinstance(sort, basestring):
-            sort = lucene.SortField(sort, lucene.SortField.STRING, reverse)
-        if not isinstance(sort, (lucene.Sort, type(None))):
-            sort = lucene.Sort(sort)
-        topdocs = super(Searcher, self).search(query, filter, count, *([sort] * bool(sort)))
-        scoredocs = list(topdocs.scoreDocs)
-        ids, scores = (map(operator.attrgetter(name), scoredocs) for name in ('doc', 'score'))
-        return Hits(self, ids, scores, topdocs.totalHits)
+        else:
+            count, inorder = min(count, self.maxDoc()), not weight.scoresDocsOutOfOrder()
+            if sort is None:
+                collector = lucene.TopScoreDocCollector.create(count, inorder)
+            else:
+                if isinstance(sort, basestring):
+                    sort = self.sorter(sort, reverse=reverse)
+                if not isinstance(sort, lucene.Sort):
+                    sort = lucene.Sort(sort)
+                collector = lucene.TopFieldCollector.create(sort, count, False, scores, maxscore, inorder)
+        TimeLimitingCollector = lucene.TimeLimitingCollector if isinstance(collector, lucene.Collector) else lucene.TimeLimitedCollector
+        results = collector if timeout is None else TimeLimitingCollector(collector, long(timeout * 1000))
+        try:
+            super(Searcher, self).search(weight, filter, results)
+        except lucene.JavaError as timeout:
+            if not TimeLimitingCollector.TimeExceededException.instance_(timeout.getJavaException()):
+                raise
+        if isinstance(collector, HitCollector):
+            ids, scores = collector.sorted(key=sort, reverse=reverse)
+            collector.finalize()
+            stats = len(ids), max(scores or [float('nan')])
+        else:
+            topdocs = collector.topDocs()
+            scoredocs = list(topdocs.scoreDocs)
+            ids, scores = (list(map(operator.attrgetter(name), scoredocs)) for name in ('doc', 'score'))
+            stats = topdocs.totalHits, topdocs.maxScore
+        stats *= not isinstance(timeout, lucene.JavaError)
+        return Hits(self, ids, scores, *stats)
     def facets(self, ids, *keys):
         """Return mapping of document counts for the intersection with each facet.
         
@@ -410,21 +463,34 @@ class Searcher(object):
             filters = self.filters.get(key)
             if isinstance(filters, lucene.Filter):
                 counts[key] = Filter.overlap(ids, filters, self.indexReader)
-            elif isinstance(key, basestring):
-                values = self.terms(key) if filters is None else filters
-                counts.update(self.facets(ids, *((key, value) for value in values)))
             else:
-                name, value = key
+                name, value = (key, None) if isinstance(key, basestring) else key
                 filters = self.filters.setdefault(name, {})
-                if value not in filters:
-                    filters[value] = Query.term(name, value).filter()
-                counts[name][value] = Filter.overlap(ids, filters[value], self.indexReader)
+                if value is None:
+                    values = filters or self.terms(name)
+                else:
+                    values = [value] if value in filters else self.terms(name, value)
+                for value in values:
+                    if value not in filters:
+                        filters[value] = Query.term(name, value).filter()
+                    counts[name][value] = Filter.overlap(ids, filters[value], self.indexReader)
         return dict(counts)
+    def sorter(self, field, type='string', parser=None, reverse=False):
+        "Return `SortField`_ with cached attributes if available."
+        sorter = self.sorters.get(field, SortField(field, type, parser, reverse))
+        return sorter if sorter.reverse == reverse else SortField(sorter.field, sorter.typename, sorter.parser, reverse)
+    def comparator(self, field, type='string', parser=None):
+        "Return :meth:`IndexReader.comparator` using a cached `SortField`_ if available."
+        return self.sorter(field, type, parser).comparator(self.indexReader)
+    def spellchecker(self, field):
+        "Return and cache spellchecker for given field."
+        try:
+            return self.spellcheckers[field]
+        except KeyError:
+            return self.spellcheckers.setdefault(field, SpellChecker(self.terms(field, counts=True)))
     def suggest(self, field, prefix, count=None):
         "Return ordered suggested words for prefix."
-        if field not in self.spellcheckers:
-            self.spellcheckers[field] = SpellChecker(self.terms(field, counts=True))
-        return self.spellcheckers[field].suggest(prefix, count)
+        return self.spellchecker(field).suggest(prefix, count)
     def correct(self, field, text, distance=2, minSimilarity=0.5):
         """Generate potential words ordered by increasing edit distance and decreasing frequency.
         For optimal performance only iterate the required slice size of corrections.
@@ -432,16 +498,15 @@ class Searcher(object):
         :param distance: the maximum edit distance to consider for enumeration
         :param minSimilarity: threshold for additional fuzzy terms after edits have been exhausted
         """
-        if field not in self.spellcheckers:
-            self.spellcheckers[field] = SpellChecker(self.terms(field, counts=True))
+        spellchecker = self.spellchecker(field)
         corrections = set()
-        for words in itertools.islice(self.spellcheckers[field].correct(text), distance + 1):
+        for words in itertools.islice(spellchecker.correct(text), distance + 1):
             for word in words:
                 yield word
             corrections.update(words)
         if minSimilarity:
             words = set(self.terms(field, text, minSimilarity=minSimilarity)) - corrections
-            for word in sorted(words, key=self.spellcheckers[field].__getitem__, reverse=True):
+            for word in sorted(words, key=spellchecker.__getitem__, reverse=True):
                 yield word
 
 class IndexSearcher(Searcher, lucene.IndexSearcher, IndexReader):
@@ -452,6 +517,15 @@ class IndexSearcher(Searcher, lucene.IndexSearcher, IndexReader):
     """
     def __init__(self, directory, analyzer=None):
         Searcher.__init__(self, *self.open(directory), analyzer=analyzer)
+    @classmethod
+    def load(cls, directory, analyzer=None):
+        "Open `Searcher`_ with a lucene RAMDirectory, loading index into memory."
+        if isinstance(directory, basestring):
+            directory = lucene.FSDirectory.open(lucene.File(directory))
+            ref = closing([directory])
+        self = cls(lucene.RAMDirectory(directory), analyzer)
+        self.shared.add(self.directory)
+        return self
 
 class MultiSearcher(Searcher, lucene.MultiSearcher, IndexReader):
     """Inherited lucene MultiSearcher.
@@ -463,10 +537,10 @@ class MultiSearcher(Searcher, lucene.MultiSearcher, IndexReader):
     def __init__(self, searchers, analyzer=None):
         if lucene.MultiReader.instance_(searchers):
             self.indexReader = searchers
-            searchers = map(lucene.IndexSearcher, searchers.sequentialSubReaders)
+            searchers = list(map(lucene.IndexSearcher, searchers.sequentialSubReaders))
         else:
             searchers = [searcher if isinstance(searcher, lucene.Searcher) else lucene.IndexSearcher(searcher) for searcher in self.open(*searchers)]
-            self.indexReader = lucene.MultiReader(map(operator.attrgetter('indexReader'), searchers), False)
+            self.indexReader = lucene.MultiReader(list(map(operator.attrgetter('indexReader'), searchers)), False)
             self.owned = closing([self.indexReader])
         Searcher.__init__(self, searchers, analyzer)
 
@@ -482,20 +556,15 @@ class IndexWriter(lucene.IndexWriter):
     :param analyzer: lucene Analyzer, default StandardAnalyzer
     """
     __len__ = lucene.IndexWriter.numDocs
-    __del__ = IndexSearcher.__del__.im_func
-    parse = Searcher.parse.im_func
+    __del__ = Searcher.__dict__['__del__']
+    parse = Searcher.__dict__['parse']
     def __init__(self, directory=None, mode='a', analyzer=None):
         self.shared = closing()
         if analyzer is None:
-            analyzer = lucene.StandardAnalyzer(lucene.Version.LUCENE_CURRENT) if hasattr(lucene, 'Version') else lucene.StandardAnalyzer()
+            analyzer = lucene.StandardAnalyzer(lucene.Version.LUCENE_CURRENT)
             self.shared.add(analyzer)
         if not isinstance(directory, lucene.Directory):
-            if directory is None:
-                directory = lucene.RAMDirectory()
-            elif hasattr(lucene.FSDirectory, 'open'):
-                directory = lucene.FSDirectory.open(lucene.File(directory))
-            else:
-                directory = lucene.FSDirectory.getDirectory(directory)
+            directory = lucene.RAMDirectory() if directory is None else lucene.FSDirectory.open(lucene.File(directory))
             self.shared.add(directory)
         mfl = lucene.IndexWriter.MaxFieldLength.LIMITED
         if mode == 'a':
@@ -506,7 +575,7 @@ class IndexWriter(lucene.IndexWriter):
     @property
     def segments(self):
         "segment filenames with document counts"
-        items = (seg.split(':c') for seg in self.segString().split())
+        items = (seg.lower().split(':c') for seg in self.segString().split())
         return dict((name, int(value)) for name, value in items)
     def set(self, name, cls=Field, **params):
         """Assign parameters to field name.
@@ -528,7 +597,7 @@ class IndexWriter(lucene.IndexWriter):
         doc.boost = boost
         for name, values in terms.items():
             if isinstance(values, Atomic):
-                values = [values] 
+                values = values,
             for field in self.fields[name].items(*values):
                 doc.add(field)
         self.addDocument(doc)
@@ -545,11 +614,7 @@ class IndexWriter(lucene.IndexWriter):
     def __iadd__(self, directory):
         "Add directory (or reader, searcher, writer) to index."
         if isinstance(directory, basestring):
-            if hasattr(lucene.FSDirectory, 'open'):
-                directory = lucene.FSDirectory.open(lucene.File(directory))
-            else:
-                directory = lucene.FSDirectory.getDirectory(directory)
-            with contextlib.closing(directory):
+            with contextlib.closing(lucene.FSDirectory.open(lucene.File(directory))) as directory:
                 self.addIndexesNoOptimize([directory])
         else:
             if not isinstance(directory, lucene.Directory):
