@@ -1,8 +1,10 @@
 """
-Wrappers for lucene NumericFields, available since version 2.9.
+Wrappers for lucene NumericFields.
 Alternative implementations of spatial and datetime fields.
 """
 
+from future_builtins import map
+import itertools
 import datetime, calendar
 import lucene
 from .documents import Field
@@ -32,16 +34,16 @@ class NumericField(Field):
         "Return lucene NumericRangeQuery."
         if isinstance(start, float) or isinstance(stop, float):
             start, stop = (value if value is None else lucene.Double(value) for value in (start, stop))
-            return lucene.NumericRangeQuery.newDoubleRange(self.name, start, stop, lower, upper)
+            return lucene.NumericRangeQuery.newDoubleRange(self.name, self.step, start, stop, lower, upper)
         if start is not None:
             start = None if start < lucene.Long.MIN_VALUE else lucene.Long(long(start))
         if stop is not None:
             stop = None if stop > lucene.Long.MAX_VALUE else lucene.Long(long(stop))
-        return lucene.NumericRangeQuery.newLongRange(self.name, start, stop, lower, upper)
+        return lucene.NumericRangeQuery.newLongRange(self.name, self.step, start, stop, lower, upper)
 
 class PointField(spatial.SpatialField, NumericField):
     __doc__ = spatial.PointField.__doc__
-    items = spatial.PointField.items.im_func
+    items = spatial.PointField.__dict__['items']
     def tiles(self, points, span=False):
         """Generate tile values from points (lng, lat).
         
@@ -57,29 +59,32 @@ class PointField(spatial.SpatialField, NumericField):
     def within(self, lng, lat, distance, limit=spatial.Tiler.base):
         """Return range queries for any tiles which could be within distance of given point.
         
-        :param lng, lat: point
+        :param lng,lat: point
         :param distance: search radius in meters
         :param limit: maximum number of tiles to consider
         """
-        queries = []
-        for tile in sorted(self.radiate(lat, lng, distance, self.precision, limit)):
-            query = self.prefix(tile)
-            if queries and queries[-1].max.equals(query.min):
-                queries[-1] = self.range(queries[-1].min.longValue(), query.max.longValue())
+        tiles = sorted(self.radiate(lat, lng, distance, self.precision, limit))
+        precision, = set(map(len, tiles))
+        shift = self.base ** (self.precision - precision)
+        slices = []
+        for tile in tiles:
+            tile = int(tile, self.base) * shift
+            if slices and slices[-1][1] == tile:
+                slices[-1] = slices[-1][0], tile + shift
             else:
-                queries.append(query)
-        return Query.any(*queries)
+                slices.append((tile, tile + shift))
+        return Query.any(*itertools.starmap(self.range, slices))
 
 class PolygonField(PointField):
     __doc__ = spatial.PolygonField.__doc__
-    items = spatial.PolygonField.items.im_func
+    items = spatial.PolygonField.__dict__['items']
 
 class DateTimeField(NumericField):
     """Field which indexes datetimes as a NumericField of timestamps.
     Supports datetimes, dates, and any prefix of time tuples.
     """
-    duration = documents.DateTimeField.duration.im_func
-    within = documents.DateTimeField.within.im_func
+    duration = documents.DateTimeField.__dict__['duration']
+    within = documents.DateTimeField.__dict__['within']
     def timestamp(self, date):
         "Return utc timestamp from date or time tuple."
         if isinstance(date, datetime.date):
