@@ -2,6 +2,7 @@
 Query wrappers and search utilities.
 """
 
+from future_builtins import filter, map
 import itertools
 import bisect
 import heapq
@@ -23,7 +24,7 @@ class Query(object):
         "Generate set of query term items."
         terms = lucene.HashSet()
         self.extractTerms(terms)
-        for term in itertools.imap(lucene.Term.cast_, terms):
+        for term in map(lucene.Term.cast_, terms):
             yield term.field(), term.text()
     @classmethod
     def term(cls, name, value):
@@ -40,13 +41,23 @@ class Query(object):
         return self
     @classmethod
     def any(cls, *queries, **terms):
-        "Return :class:`BooleanQuery` (OR) from queries and terms."
+        "Return `BooleanQuery`_ (OR) from queries and terms."
         return cls.boolean(lucene.BooleanClause.Occur.SHOULD, *queries, **terms)
     @classmethod
     def all(cls, *queries, **terms):
-        "Return :class:`BooleanQuery` (AND) from queries and terms."
+        "Return `BooleanQuery`_ (AND) from queries and terms."
         return cls.boolean(lucene.BooleanClause.Occur.MUST, *queries, **terms)
     @classmethod
+    def disjunct(cls, multiplier, *queries, **terms):
+        "Return lucene DisjunctionMaxQuery from queries and terms."
+        self = cls(lucene.DisjunctionMaxQuery, multiplier)
+        for query in queries:
+            self.add(query)
+        for name, values in terms.items():
+            for value in ([values] if isinstance(values, basestring) else values):
+                self.add(cls.term(name, value))
+        return self
+    @classmethod
     def span(cls, name, value):
         "Return :class:`SpanTermQuery <SpanQuery>`."
         return SpanQuery(lucene.SpanTermQuery, lucene.Term(name, value))
@@ -63,8 +74,7 @@ class Query(object):
     @classmethod
     def range(cls, name, start, stop, lower=True, upper=False):
         "Return lucene RangeQuery, by default with a half-open interval."
-        base = lucene.TermRangeQuery if hasattr(lucene, 'TermRangeQuery') else lucene.ConstantScoreRangeQuery
-        return cls(base, name, start, stop, lower, upper)
+        return cls(lucene.TermRangeQuery, name, start, stop, lower, upper)
     @classmethod
     def phrase(cls, name, *values):
         "Return lucene PhraseQuery.  None may be used as a placeholder."
@@ -142,7 +152,7 @@ class SpanQuery(Query):
         :param inOrder: default True
         :param collectPayloads: default True
         """
-        args = map(kwargs.pop, ('slop', 'inOrder'), (0, True))
+        args = list(map(kwargs.pop, ('slop', 'inOrder'), (0, True)))
         if 'collectPayloads' in kwargs:
             args.append(kwargs.pop('collectPayloads'))
         assert not kwargs, 'unexpected keyword arguments: {0}'.format(list(kwargs))
@@ -169,7 +179,7 @@ class HitCollector(lucene.PythonCollector if hasattr(lucene, 'PythonCollector')
         if key is None:
             key, reverse = self.scores.__getitem__, True
         ids.sort(key=key, reverse=reverse)
-        return ids, map(self.scores.__getitem__, ids)
+        return ids, list(map(self.scores.__getitem__, ids))
 
 class Filter(lucene.PythonFilter):
     "Inherited lucene Filter with a cached BitSet of ids."
@@ -180,12 +190,12 @@ class Filter(lucene.PythonFilter):
             self.docIdSet.union(ids)
         else:
             setter = self.docIdSet.set
-            for id in itertools.imap(long, ids):
+            for id in map(long, ids):
                 setter(id)
         if lucene.VERSION < '3':
             self.bitSet = lucene.BitSet()
             setter = self.bitSet.set
-            for id in itertools.ifilter(self.docIdSet.get, xrange(self.docIdSet.size())):
+            for id in filter(self.docIdSet.get, xrange(self.docIdSet.size())):
                 setter(id)
     @staticmethod
     def overlap(self, other, reader=None):
@@ -198,9 +208,30 @@ class Filter(lucene.PythonFilter):
         "Return cached OpenBitSet, reader is ignored."
         return self.docIdSet
 
+class SortField(lucene.SortField):
+    """Inherited lucene SortField used for caching FieldCache parsers.
+        
+    :param name: field name
+    :param type: type object or name compatible with SortField constants
+    :param parser: lucene FieldCache.Parser or callable applied to field values
+    :param reverse: reverse flag used with sort
+    """
+    def __init__(self, name, type='string', parser=None, reverse=False):
+        type = self.typename = getattr(type, '__name__', type).capitalize()
+        if parser is None:
+            parser = getattr(lucene.SortField, type.upper())
+        elif not lucene.FieldCache.Parser.instance_(parser):
+            base = getattr(lucene, 'Python{0}Parser'.format(type))
+            namespace = {'parse' + type: staticmethod(parser)}
+            parser = object.__class__(base.__name__, (base,), namespace)()
+        lucene.SortField.__init__(self, name, parser, reverse)
+    def comparator(self, reader):
+        "Return indexed values from default FieldCache using the given reader."
+        method = getattr(lucene.FieldCache.DEFAULT, 'get{0}s'.format(self.typename))
+        return method(reader, self.field, *([self.parser] * bool(self.parser)))
+
 class Highlighter(lucene.Highlighter):
     """Inherited lucene Filter with stored analysis options.
-    Using span scoring in lucene 2.4 is not thread-safe.
     
     :param query: lucene Query
     :param analyzer: analyzer for texts
@@ -213,10 +244,9 @@ class Highlighter(lucene.Highlighter):
     def __init__(self, query, analyzer, span=True, formatter=None, encoder=None, field=None, reader=None):
         if isinstance(formatter, basestring):
             formatter = lucene.SimpleHTMLFormatter('<{0}>'.format(formatter), '</{0}>'.format(formatter))
-        scorer = lucene.QueryScorer if (span or not hasattr(lucene, 'QueryTermScorer')) else lucene.QueryTermScorer
-        scorer = scorer(*filter(None, [query, reader, field]))
+        scorer = (lucene.QueryScorer if span else lucene.QueryTermScorer)(*filter(None, [query, reader, field]))
         lucene.Highlighter.__init__(self, *filter(None, [formatter, encoder, scorer]))
-        self.query, self.analyzer, self.span, self.field = query, analyzer, span, field
+        self.analyzer = analyzer
     def fragments(self, text, count=1, field=None):
         """Return highlighted text fragments.
         
@@ -224,12 +254,6 @@ class Highlighter(lucene.Highlighter):
         :param count: maximum number of fragments
         :param field: optional field to use for text analysis
         """
-        if lucene.VERSION <= '2.4.1': # memory leak in string array
-            tokens = self.analyzer.tokenStream(field, lucene.StringReader(text))
-            if self.span:
-                tokens = lucene.CachingTokenFilter(tokens)
-                self.fragmentScorer = lucene.HighlighterSpanScorer(self.query, self.field, tokens)
-            return map(unicode, self.getBestTextFragments(tokens, text, True, count))
         return list(self.getBestFragments(self.analyzer, field, text, count))
 
 class SpellChecker(dict):
@@ -240,7 +264,7 @@ class SpellChecker(dict):
         dict.__init__(self, *args, **kwargs)
         self.words = sorted(self)
         self.alphabet = sorted(set(itertools.chain.from_iterable(self.words)))
-        self.suffix = self.alphabet[-1] * max(itertools.imap(len, self.words)) if self.alphabet else ''
+        self.suffix = self.alphabet[-1] * max(map(len, self.words)) if self.alphabet else ''
         self.prefixes = set(word[:stop] for word in self.words for stop in range(len(word) + 1))
     def suggest(self, prefix, count=None):
         "Return ordered suggested words for prefix."
@@ -269,9 +293,9 @@ class SpellChecker(dict):
         "Generate ordered sets of words by increasing edit distance."
         previous, edits = set(), {word: 0}
         for distance in range(len(word)):
-            yield sorted(itertools.ifilter(self.__contains__, edits), key=self.__getitem__, reverse=True)
+            yield sorted(filter(self.__contains__, edits), key=self.__getitem__, reverse=True)
             previous.update(edits)
-            groups = itertools.imap(self.edits, edits, edits.values())
+            groups = map(self.edits, edits, edits.values())
             edits = dict((edit, group[edit]) for group in groups for edit in group if edit not in previous)
 
 class SpellParser(lucene.PythonQueryParser):
