@@ -8,6 +8,7 @@ See http://www.maptiler.org/google-maps-coordinates-tile-bounds-projection/.
 The quadkeys are then indexed using a prefix tree, creating a cartesian tier of tiles.
 """
 
+import warnings
 import itertools, operator
 import lucene
 from .globalmaptiles import GlobalMercator
@@ -48,13 +49,18 @@ class Tiler(GlobalMercator):
         for precision in range(precision, 0, -1):
             left, bottom = self.MetersToTile(x-distance, y-distance, precision)
             right, top = self.MetersToTile(x+distance, y+distance, precision)
-            if (right+1-left) * (top+1-bottom) <= limit:
+            count = (right+1 - left) * (top+1 - bottom)
+            if count > self.base ** precision: # spanned globe
+                for tile in range(self.base):
+                    yield str(tile)
+                return
+            if count <= limit:
                 break
         for i, j in itertools.product(range(left, right+1), range(bottom, top+1)):
             left, bottom, right, top = self.TileBounds(i, j, precision)
             dx = min(0, x-left, right-x)
             dy = min(0, y-bottom, top-y)
-            if (dx**2 + dy**2) ** 0.5 <= distance:
+            if (dx**2 + dy**2) <= distance**2:
                 yield self.QuadTree(i, j, precision)
     def zoom(self, tiles):
         "Return reduced number of tiles, by zooming out where all sub-tiles are present."
@@ -83,7 +89,7 @@ class SpatialField(Field, Tiler):
     def within(self, lng, lat, distance, limit=Tiler.base):
         """Return prefix queries for any tiles which could be within distance of given point.
         
-        :param lng, lat: point
+        :param lng,lat: point
         :param distance: search radius in meters
         :param limit: maximum number of tiles to consider
         """
@@ -103,6 +109,9 @@ class PointField(SpatialField, PrefixField):
     """Geospatial points, which create a tiered index of tiles.
     Points must still be stored if exact distances are required upon retrieval.
     """
+    def __init__(self, *args, **kwargs):
+        warnings.warn('engine.{Point,Polygon}Field will be replaced with engine.numeric.{Point,Polygon}Field.', DeprecationWarning)
+        return SpatialField.__init__(self, *args, **kwargs)
     def items(self, *points):
         "Generate tiles from points (lng, lat)."
         return super(SpatialField, self).items(*self.tiles(points))
