@@ -4,14 +4,17 @@ Restful json `CherryPy <http://cherrypy.org/>`_ server.
 CherryPy and Lucene VM integration issues:
  * Monitors (such as autoreload) are not compatible with the VM unless threads are attached.
  * WorkerThreads must be also attached to the VM.
+ * VM initialization must occur after daemonizing.
  * Recommended that the VM ignores keyboard interrupts (-Xrs) for clean server shutdown.
 """
 
+from future_builtins import filter, map
 import re
+import time
 import httplib
 import threading
-import warnings
-import itertools, collections
+import collections
+import itertools, operator
 import os, optparse
 from contextlib import contextmanager
 try:
@@ -22,27 +25,45 @@ import lucene
 import cherrypy
 import engine
 
-def json_tool():
-    "Transform responses into json format."
-    response = cherrypy.response
-    if response.status is None and response.headers['content-type'].startswith('text/'):
-        response.headers['content-type'] = 'text/x-json'
-        response.body = json.dumps(response.body)
-cherrypy.tools.json = cherrypy.Tool('before_finalize', json_tool)
+def tool(hook):
+    "Return decorator to register tool at given hook point."
+    def decorator(func):
+        setattr(cherrypy.tools, func.__name__.rstrip('_'), cherrypy.Tool(hook, func))
+        return func
+    return decorator
 
-def allow_tool(methods=['GET', 'HEAD']):
-    "Only allow specified methods."
+@tool('before_handler')
+def json_(indent=None):
+    """Transform responses into json format.
+    Specify tools.json.indent for pretty printing.
+    """
+    handler = cherrypy.request.handler
+    def json_handler(*args, **kwargs):
+        body = handler(*args, **kwargs)
+        if cherrypy.response.headers['content-type'].startswith('text/'):
+            cherrypy.response.headers['content-type'] = 'text/x-json'
+            body = json.dumps(body, indent=indent)
+        return body
+    cherrypy.request.handler = json_handler
+
+@tool('on_start_resource')
+def allow(methods=('GET', 'HEAD')):
+    "Only allow methods specified in tools.allow.methods."
     request = cherrypy.request
     if request.method not in methods and not isinstance(request.handler, cherrypy.HTTPError):
         cherrypy.response.headers['allow'] = ', '.join(methods)
         message = "The path {0!r} does not allow {1}.".format(request.path_info, request.method)
         raise cherrypy.HTTPError(httplib.METHOD_NOT_ALLOWED, message)
-cherrypy.tools.allow = cherrypy.Tool('on_start_resource', allow_tool)
+
+@tool('before_finalize')
+def time_():
+    "Return response time in headers."
+    cherrypy.response.headers['x-response-time'] = time.time() - cherrypy.response.time
 
 def json_error(version, **body):
     "Transform errors into json format."
     cherrypy.response.headers['content-type'] = 'text/x-json'
-    return json.dumps(body)
+    return json.dumps(body, indent=2)
 
 def attach_thread(id):
     "Attach current cherrypy worker thread to lucene VM."
@@ -54,14 +75,14 @@ class Autoreloader(cherrypy.process.plugins.Autoreloader):
         lucene.getVMEnv().attachCurrentThread()
         cherrypy.process.plugins.Autoreloader.run(self)
 
-class Autorefresher(cherrypy.process.plugins.Monitor):
-    "Automatically refresh WebSearcher."
-    def __init__(self, bus, root, frequency):
+class AttachedMonitor(cherrypy.process.plugins.Monitor):
+    "Periodically run a callback function in an attached thread."
+    def __init__(self, bus, callback, frequency=cherrypy.process.plugins.Monitor.frequency):
         cherrypy.process.plugins.Monitor.__init__(self, bus, self.run, frequency)
-        self.method = getattr(root, 'commit', root.refresh)
+        self._callback = callback
     def run(self):
         lucene.getVMEnv().attachCurrentThread()
-        self.method()
+        self._callback()
 
 @contextmanager
 def HTTPError(status, *exceptions):
@@ -73,7 +94,7 @@ def HTTPError(status, *exceptions):
 
 class WebSearcher(object):
     "Dispatch root with a delegated Searcher."
-    _cp_config = {'tools.json.on': True, 'tools.allow.on': True, 'error_page.default': json_error,
+    _cp_config = {'tools.json.on': True, 'tools.allow.on': True, 'tools.time.on': True, 'error_page.default': json_error,
         'tools.gzip.on': True, 'tools.gzip.mime_types': ['text/html', 'text/plain', 'text/x-json']}
     def __init__(self, *directories, **kwargs):
         self.indexer = engine.MultiSearcher(directories, **kwargs) if len(directories) > 1 else engine.IndexSearcher(*directories, **kwargs)
@@ -91,16 +112,34 @@ class WebSearcher(object):
             (field, boost), = fields
         else:
             field = [name for name, boost in fields] or ''
+        if 'type' in options:
+            with HTTPError(httplib.BAD_REQUEST, AttributeError):
+                return getattr(engine.Query, options['type'])(field, q)
         for key in set(options) - set(['op', 'version']):
             with HTTPError(httplib.BAD_REQUEST, ValueError):
                 options[key] = json.loads(options[key])
         return q and self.indexer.parse(q, field=field, **options)
+    def select(self, fields=None, **options):
+        "Return parsed field selectors: stored, multi-valued, and indexed."
+        if fields is not None:
+            fields = dict.fromkeys(filter(None, fields.split(',')))
+        if 'multifields' in options:
+            options['fields.multi'] = options.pop('multifields')
+            cherrypy.response.headers['Warning'] = '199 lupyne "multifields has been deprecated and renamed fields.multi"'
+        multi = list(filter(None, options.get('fields.multi', '').split(',')))
+        indexed = [field.split(':') for field in options.get('fields.indexed', '').split(',') if field]
+        return fields, multi, indexed
     @cherrypy.expose
     @cherrypy.tools.allow(methods=['POST'])
     def refresh(self, **caches):
-        """Reopen searcher.
+        """Refresh index.
         
         **POST** /refresh
+            Reopen searcher, optionally reloading caches, and return document count.
+            
+            &filters&sorters&spellcheckers
+            
+            :return: *int*
         """
         self.indexer = self.indexer.reopen(**dict.fromkeys(caches, True))
         return len(self.indexer)
@@ -113,13 +152,11 @@ class WebSearcher(object):
             
             :return: {*string*: *int*}
         """
-        if not isinstance(self.indexer, lucene.MultiSearcher):
-            return {str(self.indexer.directory): len(self.indexer)}
-        if hasattr(lucene.MultiReader, 'sequentialSubReaders'):
-            return dict((str(reader.directory()), reader.numDocs()) for reader in self.indexer.sequentialSubReaders)
-        return {str(self.indexer): len(self.indexer)}
+        if isinstance(self.indexer, lucene.MultiSearcher):
+            return dict((unicode(reader.directory()), reader.numDocs()) for reader in self.indexer.sequentialSubReaders)
+        return {unicode(self.indexer.directory): len(self.indexer)}
     @cherrypy.expose
-    def docs(self, id=None, fields=None, multifields=''):
+    def docs(self, id=None, fields=None, **options):
         """Return ids or documents.
         
         **GET** /docs
@@ -128,48 +165,50 @@ class WebSearcher(object):
             :return: [*int*,... ]
         
         **GET** /docs/*int*?
-            Return document mappings, optionally selected unique or multi-valued fields.
+            Return document mappings, optionally selecting stored, multi-valued, and cached indexed fields.
             
-            &fields=\ *chars*,...
-            
-            &multifields=\ *chars*,...
+            &fields=\ *chars*,... &fields.multi=\ *chars*,... &fields.indexed=\ *chars*\ [:*chars*],...
             
             :return: {*string*: *string*\|\ *array*,... }
         """
         if id is None:
             return list(self.indexer)
-        with HTTPError(httplib.NOT_FOUND, ValueError):
+        fields, multi, indexed = self.select(fields, **options)
+        with HTTPError(httplib.NOT_FOUND, ValueError, lucene.JavaError):
             id = int(id)
-        if fields is not None:
-            fields = dict.fromkeys(filter(None, fields.split(',')))
-        multifields = filter(None, multifields.split(','))
-        with HTTPError(httplib.NOT_FOUND, lucene.JavaError):
-            doc = self.indexer[id] if fields is None else self.indexer.get(id, *itertools.chain(fields, multifields))
-        return doc.dict(*multifields, **(fields or {}))
+            doc = self.indexer[id] if fields is None else self.indexer.get(id, *itertools.chain(fields, multi))
+        result = doc.dict(*multi, **(fields or {}))
+        result.update((item[0], self.indexer.comparator(*item)[id]) for item in indexed)
+        return result
     @cherrypy.expose
-    def search(self, q=None, count=None, start=0, fields=None, multifields='', sort=None, facets='', hl='', mlt=None, spellcheck=0, **options):
+    def search(self, q=None, count=None, start=0, fields=None, sort=None, facets='', group='', hl='', mlt=None, spellcheck=0, timeout=None, **options):
         """Run query and return documents.
         
         **GET** /search?
             Return list of document objects and total doc count.
             
-            &q=\ *chars*\ &q.\ *chars*\ =...,
-                query and optional parser settings: q.field, q.op,...
+            &q=\ *chars*\ &q.type=[term|prefix|wildcard]&q.\ *chars*\ =...,
+                query, optional type to skip parsing, and optional parser settings: q.field, q.op,...
+            
+            &filter=\ *chars*
+                | cached filter applied to the query
+                | if a previously cached filter is not found, the value will be parsed as a query
             
             &count=\ *int*\ &start=0
                 maximum number of docs to return and offset to start at
             
-            &fields=\ *chars*,...
-                only include selected fields
+            &fields=\ *chars*,... &fields.multi=\ *chars*,... &fields.indexed=\ *chars*\ [:*chars*],...
+                only include selected stored fields; multi-valued fields returned in an array; indexed fields are cached
             
-            &multifields=\ *chars*,...
-                multi-valued fields returned in an array
-            
-            &sort=\ [-]\ *chars*\ [:*chars*],...
-                field name, optional type, minus sign indicates descending
+            &sort=\ [-]\ *chars*\ [:*chars*],... &sort.scores[=max]
+                | field name, optional type, minus sign indicates descending
+                | optionally score docs, additionally compute maximum score
             
             &facets=\ *chars*,...
-                include facet counts for given field names;  facets filters are cached
+                include facet counts for given field names; facets filters are cached
+            
+            &group=\ *chars*\ &group.count=1
+                group documents by field value, up to given maximum count
             
             &hl=\ *chars*,... &hl.count=1&hl.tag=strong&hl.enable=[fields|terms]
                 | stored fields to return highlighted
@@ -183,14 +222,19 @@ class WebSearcher(object):
             
             &spellcheck=\ *int*
                 | maximum number of spelling corrections to return for each query term, grouped by field
-                | original query is still run;  use q.spellcheck=true to affect query parsing
+                | original query is still run; use q.spellcheck=true to affect query parsing
+            
+            &timeout=\ *float*
+                timeout search after elapsed number of seconds
             
             :return:
                 | {
                 | "query": *string*,
-                | "count": *int*,
-                | "docs": [{"__id__": *int*, "__score__": *number*, "__highlights__": {*string*: *array*,... }, *string*: *string*\|\ *array*,... },... ],
+                | "count": *int*\|null,
+                | "maxscore": *float*\|null,
+                | "docs": [{"__id__": *int*, "__score__": *float*, "__highlights__": {*string*: *array*,... }, *string*: *string*\|\ *array*,... },... ],
                 | "facets": {*string*: {*string*: *int*,... },... },
+                | "groups": [{"count": *int*, "value": *value*, "docs": [{... },... ]},... ]
                 | "spellcheck": {*string*: {*string*: [*string*,... ],... },... },
                 | }
         """
@@ -199,22 +243,31 @@ class WebSearcher(object):
             if count is not None:
                 count = int(count) + start
             spellcheck = int(spellcheck)
+            if timeout is not None:
+                timeout = float(timeout)
+            gcount = int(options.get('group.count', 1))
+            hlcount = int(options.get('hl.count', 1))
+            if mlt is not None:
+                mlt = int(mlt)
         reverse = False
         searcher = getattr(self.indexer, 'indexSearcher', self.indexer)
         if sort is not None:
             sort = (re.match('(-?)(\w+):?(\w*)', field).groups() for field in sort.split(','))
-            sort = [(name, (type.upper() or 'STRING'), (reverse == '-')) for reverse, name, type in sort]
+            sort = [(name, (type or 'string'), (reverse == '-')) for reverse, name, type in sort]
             if count is None:
-                with HTTPError(httplib.BAD_REQUEST, ValueError):
-                    (name, type, reverse), = sort # only one sort field allowed with unlimited count
-                sort = searcher.comparator(name, type).__getitem__
+                with HTTPError(httplib.BAD_REQUEST, ValueError, AttributeError):
+                    reverse, = set(reverse for name, type, reverse in sort) # only one sort direction allowed with unlimited count
+                    comparators = [searcher.comparator(name, type) for name, type, reverse in sort]
+                sort = comparators[0].__getitem__ if len(comparators) == 1 else lambda id: tuple(map(operator.itemgetter(id), comparators))
             else:
                 with HTTPError(httplib.BAD_REQUEST, AttributeError):
-                    sort = [lucene.SortField(name, getattr(lucene.SortField, type), reverse) for name, type, reverse in sort]
+                    sort = [searcher.sorter(name, type, reverse=reverse) for name, type, reverse in sort]
         q = self.parse(q, **options)
+        qfilter = options.pop('filter', None)
+        if qfilter is not None and qfilter not in searcher.filters:
+            searcher.filters[qfilter] = engine.Query.__dict__['filter'](self.parse(qfilter, **options))
+        qfilter = searcher.filters.get(qfilter)
         if mlt is not None:
-            with HTTPError(httplib.BAD_REQUEST, ValueError):
-                mlt = int(mlt)
             if q is not None:
                 mlt = searcher.search(q, count=mlt+1, sort=sort, reverse=reverse).ids[mlt]
             mltfields = filter(None, options.pop('mlt.fields', '').split(','))
@@ -222,32 +275,54 @@ class WebSearcher(object):
                 attrs = dict((key.partition('.')[-1], json.loads(options[key])) for key in options if key.startswith('mlt.'))
             q = searcher.morelikethis(mlt, *mltfields, **attrs)
         if count == 0:
-            hits = searcher.search(q, count=1, sort=sort, reverse=reverse)[:0]
-        else:
-            hits = searcher.search(q, count=count, sort=sort, reverse=reverse)
-        result = {'query': q and unicode(q), 'count': hits.count, 'docs': []}
+            start = count = 1
+        scores = options.get('sort.scores')
+        scores = {'scores': scores is not None, 'maxscore': scores == 'max'}
+        hits = searcher.search(q, filter=qfilter, count=count, sort=sort, reverse=reverse, timeout=timeout, **scores)[start:]
+        result = {'query': q and unicode(q), 'count': hits.count, 'maxscore': hits.maxscore}
         tag = options.get('hl.tag', 'strong')
         field = 'fields' not in options.get('hl.enable', '') or None
         span = 'terms' not in options.get('hl.enable', '')
         if hl:
             hl = dict((name, searcher.highlighter(q, span=span, field=(field and name), formatter=tag)) for name in hl.split(','))
-        with HTTPError(httplib.BAD_REQUEST, ValueError):
-            count = int(options.get('hl.count', 1))
-        if fields is not None:
-            fields = dict.fromkeys(filter(None, fields.split(',')))
-            hits.fields = lucene.MapFieldSelector(list(itertools.chain(fields, multifields, hl)))
-        fields = fields or {}
-        multifields = filter(None, multifields.split(','))
-        for hit in hits[start:]:
-            doc = hit.dict(*multifields, **fields)
-            result['docs'].append(doc)
+        fields, multi, indexed = self.select(fields, **options)
+        if fields is None:
+            fields = {}
+        else:
+            hits.fields = lucene.MapFieldSelector(list(itertools.chain(fields, multi, hl)))
+            fields = fields or {'__id__': None}
+        indexed = dict((item[0], searcher.comparator(*item)) for item in indexed)
+        docs = []
+        groups = collections.defaultdict(lambda: {'docs': [], 'count': 0, 'index': len(groups)})
+        if group:
+            group = searcher.comparator(*group.split(':'))
+            ids, scores = [], []
+            for id, score in hits.items():
+                item = groups[group[id]]
+                item['count'] += 1
+                if item['count'] <= gcount:
+                    ids.append(id)
+                    scores.append(score)
+            hits.ids, hits.scores = ids, scores
+        for hit in hits:
+            doc = hit.dict(*multi, **fields)
+            doc.update((name, indexed[name][hit.id]) for name in indexed)
             if hl:
-                doc['__highlights__'] = dict((name, hl[name].fragments(hit[name], count)) for name in hl if name in hit)
+                doc['__highlights__'] = dict((name, hl[name].fragments(hit[name], hlcount)) for name in hl if name in hit)
+            (groups[group[hit.id]]['docs'] if group else docs).append(doc)
+        for name in groups:
+            groups[name]['value'] = name
+        if group:
+            result['groups'] = sorted(groups.values(), key=lambda item: item.pop('index'))
+        else:
+            result['docs'] = docs
+        q = q or lucene.MatchAllDocsQuery()
         if facets:
-            result['facets'] = searcher.facets(engine.Query.filter.im_func(q), *facets.split(','))
+            facets = (tuple(facet.split(':')) if ':' in facet else facet for facet in facets.split(','))
+            result['facets'] = searcher.facets(engine.Query.__dict__['filter'](q), *facets)
         if spellcheck:
             terms = result['spellcheck'] = collections.defaultdict(dict)
-            for name, value in engine.Query.terms.im_func(q):
+            for name, value in engine.Query.__dict__['terms'](q):
                 terms[name][value] = list(itertools.islice(searcher.correct(name, value), spellcheck))
         return result
     @cherrypy.expose
@@ -273,7 +348,7 @@ class WebSearcher(object):
         
         **GET** /terms/*chars*/*chars*\[\*\|~\]?count=\ *int*
             Return spellchecked term values ordered by decreasing document frequency.
-            Prefixes (*) are optimized to be suitable for real-time query suggestions;  all terms are cached.
+            Prefixes (*) are optimized to be suitable for real-time query suggestions; all terms are cached.
             
             :return: [*string*,... ]
         
@@ -303,7 +378,8 @@ class WebSearcher(object):
             start, stop = value.split(':')
             return list(self.indexer.terms(name, start, stop or None))
         if 'count' in options:
-            count = int(options['count'])
+            with HTTPError(httplib.BAD_REQUEST, ValueError):
+                count = int(options['count'])
             if value.endswith('*'):
                 return self.indexer.suggest(name, value.rstrip('*'), count)
             if value.endswith('~'):
@@ -332,20 +408,21 @@ class WebIndexer(WebSearcher):
     @cherrypy.expose
     @cherrypy.tools.allow(methods=['POST'])
     def refresh(self, **caches):
-        raise cherrypy.HTTPRedirect('/commit', httplib.MOVED_PERMANENTLY)
+        raise cherrypy.HTTPRedirect(cherrypy.request.script_name + '/commit', httplib.MOVED_PERMANENTLY)
     @cherrypy.expose
     @cherrypy.tools.allow(methods=['POST'])
     def commit(self, **caches):
-        """Commit write operations.
+        """Commit index changes.
         
         **POST** /commit
+            Commit write operations and return document count.  See :meth:`WebSearcher.refresh` for caching options.
         """
         with self.lock:
             self.indexer.commit(**dict.fromkeys(caches, True))
         return len(self.indexer)
     @cherrypy.expose
     @cherrypy.tools.allow(methods=['GET', 'HEAD', 'POST'])
-    def docs(self, id=None, fields=None, multifields='', docs='[]'):
+    def docs(self, id=None, docs='[]', fields=None, **options):
         """Add or return documents.  See :meth:`WebSearcher.docs` for GET method.
         
         **POST** /docs
@@ -354,7 +431,7 @@ class WebIndexer(WebSearcher):
             docs=[{*string*: *string*\|\ *array*,... },... ]
         """
         if cherrypy.request.method != 'POST':
-            return WebSearcher.docs(self, id, fields, multifields)
+            return WebSearcher.docs(self, id, fields, **options)
         with HTTPError(httplib.BAD_REQUEST, ValueError):
             docs = json.loads(docs)
         for doc in docs:
@@ -401,11 +478,16 @@ class WebIndexer(WebSearcher):
             field = self.indexer.fields[name]
         return dict((name, str(getattr(field, name))) for name in ['store', 'index', 'termvector'])
 
-def start(root, path='', config=None, pidfile='', daemonize=False, autoreload=0, autorefresh=0):
+def start(root=None, path='', config=None, pidfile='', daemonize=False, autoreload=0, autorefresh=0, callback=None):
     """Attach root, subscribe to plugins, and start server.
-    See cherrypy.quickstart and command-line options for documentation."""
+    
+    :param root,path,config: see cherrypy.quickstart
+    :param pidfile,daemonize,autoreload,autorefresh: see command-line options
+    :param callback: optional callback function scheduled after daemonizing
+    """
     cherrypy.engine.subscribe('start_thread', attach_thread)
-    cherrypy.engine.subscribe('stop', root.close)
+    if hasattr(root, 'close'):
+        cherrypy.engine.subscribe('stop', root.close)
     cherrypy.config['engine.autoreload.on'] = False
     if pidfile:
         cherrypy.process.plugins.PIDFile(cherrypy.engine, os.path.abspath(pidfile)).subscribe()
@@ -415,13 +497,12 @@ def start(root, path='', config=None, pidfile='', daemonize=False, autoreload=0,
     if autoreload:
         Autoreloader(cherrypy.engine, autoreload).subscribe()
     if autorefresh:
-        Autorefresher(cherrypy.engine, root, autorefresh).subscribe()
+        AttachedMonitor(cherrypy.engine, getattr(root, 'commit', root.refresh), autorefresh).subscribe()
+    if callback:
+        priority = (cherrypy.process.plugins.Daemonizer.start.priority + cherrypy.server.start.priority) // 2
+        cherrypy.engine.subscribe('start', callback, priority)
     cherrypy.quickstart(root, path, config)
 
-def main(root, path='', config=None):
-    warnings.warn("'main' function has been renamed 'start'.", DeprecationWarning)
-    start(root, path, config)
-
 parser = optparse.OptionParser(usage='python %prog [index_directory ...]')
 parser.add_option('-r', '--read-only', action='store_true', help='expose only read methods; no write lock')
 parser.add_option('-c', '--config', help='optional configuration file or json object of global params')
@@ -432,10 +513,12 @@ parser.add_option('--autorefresh', type=int, metavar='SECONDS', help='automatica
 
 if __name__ == '__main__':
     options, args = parser.parse_args()
-    if lucene.getVMEnv() is None:
-        lucene.initVM(lucene.CLASSPATH, vmargs='-Xrs')
     read_only = options.__dict__.pop('read_only')
-    root = (WebSearcher if (read_only or len(args) > 1) else WebIndexer)(*args)
     if options.config and not os.path.exists(options.config):
         options.config = {'global': json.loads(options.config)}
-    start(root, **options.__dict__)
+    directories = list(map(os.path.abspath, args))
+    root = object.__new__(WebSearcher if (read_only or len(args) > 1) else WebIndexer)
+    def init():
+        lucene.initVM(vmargs='-Xrs')
+        root.__init__(*directories)
+    start(root, callback=init, **options.__dict__)
