@@ -1,3 +1,4 @@
+from future_builtins import map
 import unittest
 import heapq
 import socket, httplib
@@ -6,10 +7,10 @@ import local, remote
 
 class TestCase(remote.BaseTest):
     ports = 8080, 8081, 8082
-    hosts = map('localhost:{0:d}'.format, ports)
+    hosts = list(map('localhost:{0:d}'.format, ports))
     def setUp(self):
         local.BaseTest.setUp(self)
-        self.servers = map(self.start, self.ports)
+        self.servers = list(map(self.start, self.ports))
     
     def testInterface(self):
         "Distributed reading and writing."
@@ -27,9 +28,9 @@ class TestCase(remote.BaseTest):
         assert all(response() == {'index': 'NOT_ANALYZED', 'store': 'YES', 'termvector': 'NO'} for response in responses)
         doc = {'name': 'sample', 'text': 'hello world'}
         responses = resources.broadcast('POST', '/docs', {'docs': [doc]})
-        assert all(response() == '' for response in responses)
+        assert all(response() is None for response in responses)
         response = resources.unicast('POST', '/docs', {'docs': [doc]})
-        assert response() == ''
+        assert response() is None
         responses = resources.broadcast('POST', '/commit')
         assert all(response() >= 1 for response in responses)
         responses = resources.broadcast('GET', '/search?q=text:hello')
