@@ -1,3 +1,4 @@
+from future_builtins import map, zip
 import unittest
 import os
 import tempfile, shutil
@@ -10,20 +11,18 @@ import lucene
 from lupyne import engine
 import fixture
 
-numeric = hasattr(lucene, 'NumericField')
-
 if issubclass(lucene.TokenFilter, collections.Iterable):
     def typeAsPayload(tokens):
         "Generator variant of lucene TypeAsPayloadTokenFilter."
         for token in tokens:
-            token.payload = lucene.Payload(lucene.JArray_byte(bytes(token.type())))
+            token.payload = lucene.Payload(lucene.JArray_byte(token.type().encode('utf8')))
             yield token
 else:
     class typeAsPayload(engine.TokenFilter):
         "Custom implementation of lucene TypeAsPayloadTokenFilter."
         def incrementToken(self):
             result = engine.TokenFilter.incrementToken(self)
-            self.payload = bytes(self.type)
+            self.payload = self.type.encode('utf8')
             return result
 
 class BaseTest(unittest.TestCase):
@@ -37,7 +36,7 @@ class TestCase(BaseTest):
     def testInterface(self):
         "Indexer and document interfaces."
         self.assertRaises(TypeError, engine.IndexSearcher)
-        analyzer = lucene.StandardAnalyzer(lucene.Version.LUCENE_CURRENT) if hasattr(lucene, 'Version') else lucene.StandardAnalyzer()
+        analyzer = lucene.StandardAnalyzer(lucene.Version.LUCENE_CURRENT)
         stemmer = engine.Analyzer(analyzer, lucene.PorterStemFilter, typeAsPayload)
         token = stemmer.tokens('hello').next()
         assert token.positionIncrement == 1
@@ -47,8 +46,7 @@ class TestCase(BaseTest):
             assert token.type == token.payload == '<ALPHANUM>'
             assert token.offset == (0, 5)
         assert str(stemmer.parse('hellos', field=['body', 'title'])) == 'body:hello title:hello'
-        if lucene.VERSION >= '2.9':
-            assert str(stemmer.parse('hellos', field={'body': 1.0, 'title': 2.0})) == 'body:hello title:hello^2.0'
+        assert str(stemmer.parse('hellos', field={'body': 1.0, 'title': 2.0})) == 'body:hello title:hello^2.0'
         indexer = engine.Indexer(analyzer=stemmer)
         self.assertRaises(lucene.JavaError, engine.Indexer, indexer.directory)
         indexer.set('text')
@@ -105,17 +103,18 @@ class TestCase(BaseTest):
         assert indexer.search('text:*hello', allowLeadingWildcard=True)
         query = engine.Query.span('text', 'world')
         self.assertRaises(AssertionError, query.near, option=None)
-        if lucene.VERSION >= '2.9':
-            assert str(query.mask('name')) == 'mask(text:world) as name'
-            query = engine.Query.near('text', 'hello', ('tag', 'python'), slop=-1, inOrder=False)
-            assert str(query) == 'spanNear([text:hello, mask(tag:python) as text], -1, false)' and indexer.count(query) == 1
-            query = engine.Query.near('text', 'hello', 'world')
-            (doc, items), = indexer.spans(query, payloads=True)
-            (start, stop, payloads), = items
-            assert doc == 0 and start == 0 and stop == 2 and payloads == ['<ALPHANUM>', '<ALPHANUM>']
-            query = engine.Query.near('text', 'hello', 'world', collectPayloads=False)
-            (doc, items), = indexer.spans(query, payloads=True)
-            assert doc == 0 and items == []
+        assert str(query.mask('name')) == 'mask(text:world) as name'
+        query = engine.Query.disjunct(0.1, query, name='sample')
+        assert str(query) == '(text:world | name:sample)~0.1'
+        query = engine.Query.near('text', 'hello', ('tag', 'python'), slop=-1, inOrder=False)
+        assert str(query) == 'spanNear([text:hello, mask(tag:python) as text], -1, false)' and indexer.count(query) == 1
+        query = engine.Query.near('text', 'hello', 'world')
+        (doc, items), = indexer.spans(query, payloads=True)
+        (start, stop, payloads), = items
+        assert doc == 0 and start == 0 and stop == 2 and payloads == ['<ALPHANUM>', '<ALPHANUM>']
+        query = engine.Query.near('text', 'hello', 'world', collectPayloads=False)
+        (doc, items), = indexer.spans(query, payloads=True)
+        assert doc == 0 and items == []
         indexer.delete('name:sample')
         indexer.delete('tag', 'python')
         assert 0 in indexer
@@ -145,6 +144,8 @@ class TestCase(BaseTest):
         for doc in fixture.constitution.docs():
             indexer.add(doc, boost=('article' in doc) + 1.0)
         indexer.commit()
+        searcher = engine.IndexSearcher.load(self.tempdir)
+        assert len(indexer) == len(searcher) and lucene.RAMDirectory.instance_(searcher.directory)
         assert indexer.filters == indexer.spellcheckers == {}
         assert indexer.facets([], 'amendment') and indexer.suggest('amendment', '')
         assert list(indexer.filters) == list(indexer.spellcheckers) == ['amendment']
@@ -174,7 +175,7 @@ class TestCase(BaseTest):
         counts = docs.values()
         assert len(docs) == count and all(counts) and sum(counts) > count
         positions = dict(indexer.positions('text', 'people'))
-        assert map(len, positions.values()) == counts
+        assert list(map(len, positions.values())) == counts
         hit, = indexer.search('"We the People"', field='text')
         assert hit['article'] == 'Preamble'
         assert sorted(hit.dict()) == ['__id__', '__score__', 'article']
@@ -182,20 +183,26 @@ class TestCase(BaseTest):
         assert hits[0]['article'] == 'Preamble'
         assert len(hits) == hits.count == 8
         assert set(map(type, hits.ids)) == set([int]) and set(map(type, hits.scores)) == set([float])
+        assert hits.maxscore == max(hits.scores)
         ids = hits.ids
         hits = indexer.search('people', count=5, field='text')
         assert hits.ids == ids[:len(hits)]
         assert len(hits) == 5 and hits.count == 8
         assert not any(map(math.isnan, hits.scores))
-        hits = indexer.search('text:people', count=5, sort=lucene.SortField('amendment', lucene.SortField.INT))
-        assert [hit.get('amendment') for hit in hits] == [None, None, '1', u'2', u'4']
-        if lucene.VERSION >= '2.9':
-            assert all(map(math.isnan, hits.scores))
-        parser = (lambda value: int(value or -1)) if hasattr(lucene, 'PythonIntParser') else None
-        comparator = indexer.comparator('amendment', type=int, parser=parser)
+        assert hits.maxscore == max(hits.scores)
+        sort = engine.SortField('amendment', type=int)
+        hits = indexer.search('text:people', count=5, sort=sort)
+        assert [hit.get('amendment') for hit in hits] == [None, None, '1', '2', '4']
+        assert all(map(math.isnan, hits.scores))
+        hits = indexer.search('text:right', count=10**7, sort=sort, scores=True)
+        assert not any(map(math.isnan, hits.scores)) and sorted(hits.scores, reverse=True) != hits.scores
+        assert math.isnan(hits.maxscore)
+        hits = indexer.search('text:right', count=2, sort=sort, maxscore=True)
+        assert hits.maxscore > max(hits.scores)
+        comparator = indexer.comparator('amendment', type=int, parser=lambda value: int(value or -1))
         hits = indexer.search('text:people', sort=comparator.__getitem__)
         assert sorted(hits.ids) == hits.ids and hits.ids != ids
-        comparator = zip(*map(indexer.comparator, ['article', 'amendment']))
+        comparator = list(zip(*map(indexer.comparator, ['article', 'amendment'])))
         hits = indexer.search('text:people', sort=comparator.__getitem__)
         assert sorted(hits.ids) != hits.ids
         hits = indexer.search('text:people', count=5, sort='amendment', reverse=True)
@@ -254,12 +261,10 @@ class TestCase(BaseTest):
         near = engine.Query.near('text', 'persons', 'papers', slop=2)
         (id, positions), = indexer.spans(near, positions=True)
         assert indexer[id]['amendment'] == '4' and positions in ([(3, 6)], [(10, 13)])
-        with warnings.catch_warnings(record=True) as leaks:
-            assert 'persons' in indexer.termvector(id, 'text')
-            assert dict(indexer.termvector(id, 'text', counts=True))['persons'] == 2
-            assert dict(indexer.positionvector(id, 'text'))['persons'] in ([3, 26], [10, 48])
-            assert dict(indexer.positionvector(id, 'text', offsets=True))['persons'] == [(46, 53), (301, 308)]
-        assert len(leaks) == 2 * (lucene.VERSION <= '2.4.1')
+        assert 'persons' in indexer.termvector(id, 'text')
+        assert dict(indexer.termvector(id, 'text', counts=True))['persons'] == 2
+        assert dict(indexer.positionvector(id, 'text'))['persons'] in ([3, 26], [10, 48])
+        assert dict(indexer.positionvector(id, 'text', offsets=True))['persons'] == [(46, 53), (301, 308)]
         query = indexer.morelikethis(0)
         assert str(query) == 'text:united text:states'
         hits = indexer.search(query & engine.Query.prefix('article', ''))
@@ -295,11 +300,11 @@ class TestCase(BaseTest):
         for name, params in fixture.zipcodes.fields.items():
             indexer.set(name, **params)
         indexer.set('longitude', engine.PrefixField, store=True)
-        indexer.fields['location'] = engine.NestedField('state:county:city')
+        indexer.fields['location'] = engine.NestedField('state.county.city', sep='.')
         for doc in fixture.zipcodes.docs():
             if doc['state'] in ('CA', 'AK', 'WY', 'PR'):
                 lat, lng = ('{0:08.3f}'.format(doc.pop(l)) for l in ['latitude', 'longitude'])
-                location = ':'.join(doc[name] for name in ['state', 'county', 'city'])
+                location = '.'.join(doc[name] for name in ['state', 'county', 'city'])
                 indexer.add(doc, latitude=lat, longitude=lng, location=location)
         indexer.commit()
         assert set(['state', 'zipcode']) < set(indexer.names('indexed'))
@@ -313,13 +318,13 @@ class TestCase(BaseTest):
         assert hit['state'] == 'PR' and hit['county'] == 'Culebra'
         states = list(indexer.terms('state'))
         assert states[0] == 'AK' and states[-1] == 'WY'
-        counties = [term.split(':')[-1] for term in indexer.terms('state:county', 'CA', 'CA~')]
+        counties = [term.split('.')[-1] for term in indexer.terms('state.county', 'CA', 'CA~')]
         field = indexer.fields['location']
         hits = indexer.search(field.prefix('CA'))
         assert sorted(set(hit['county'] for hit in hits)) == counties
         assert counties[0] == 'Alameda' and counties[-1] == 'Yuba'
-        cities = [term.split(':')[-1] for term in indexer.terms('state:county:city', 'CA:Los Angeles', 'CA:Los Angeles~')]
-        hits = indexer.search(field.prefix('CA:Los Angeles'))
+        cities = [term.split('.')[-1] for term in indexer.terms('state.county.city', 'CA.Los Angeles', 'CA.Los Angeles~')]
+        hits = indexer.search(field.prefix('CA.Los Angeles'))
         assert sorted(set(hit['city'] for hit in hits)) == cities
         assert cities[0] == 'Acton' and cities[-1] == 'Woodland Hills'
         hit, = indexer.search('zipcode:90210')
@@ -335,21 +340,41 @@ class TestCase(BaseTest):
         assert count == indexer.count(field.range(lng[:3], lng[:3]+'~'))
         assert count > indexer.count(engine.Query.term('state', 'CA'), filter=engine.Query.term(longitude, lng).filter())
         query = engine.Query.prefix('zipcode', '90')
-        (field, facets), = indexer.facets(indexer.search(query).ids, 'state:county').items()
-        assert [facets] == indexer.facets(query.filter(), 'state:county').values()
-        assert field == 'state:county'
+        (field, facets), = indexer.facets(indexer.search(query).ids, 'state.county').items()
+        assert [facets] == indexer.facets(query.filter(), 'state.county').values()
+        assert field == 'state.county'
         la, orange = sorted(filter(facets.get, facets))
-        assert la == 'CA:Los Angeles' and facets[la] > 100
-        assert orange == 'CA:Orange' and facets[orange] > 10
+        assert la == 'CA.Los Angeles' and facets[la] > 100
+        assert orange == 'CA.Orange' and facets[orange] > 10
+        (field, facets), = indexer.facets(query.filter(), ('state.county', 'CA.*')).items()
+        assert all(value.startswith('CA.') for value in facets) and set(facets) < set(indexer.filters['state.county'])
+        for count in (None, len(indexer)):
+            hits = indexer.search(query, count=count, timeout=0.01)
+            assert 0 <= len(hits) <= indexer.count(query) and hits.count in (None, len(hits)) and hits.maxscore in (None, 1.0)
+            hits = indexer.search(query, count=count, timeout=-1)
+            assert len(hits) == 0 and hits.count is hits.maxscore is None
+        directory = lucene.RAMDirectory()
+        query = engine.Query.term('state', 'CA')
+        size = indexer.copy(directory, query)
+        searcher = engine.IndexSearcher(directory)
+        assert len(searcher) == size and list(searcher.terms('state')) == ['CA']
+        path = os.path.join(self.tempdir, 'temp')
+        size = indexer.copy(path, exclude=query, optimize=True)
+        assert len(searcher) + size == len(indexer)
+        searcher = engine.IndexSearcher(path)
+        assert searcher.optimized and 'CA' not in searcher.terms('state')
+        directory.close()
     
     def testSpatial(self):
         "Spatial tile test."
-        for PolygonField in [engine.PolygonField] + [engine.numeric.PolygonField] * numeric:
+        for PolygonField in (engine.PolygonField, engine.numeric.PolygonField):
             PointField, = PolygonField.__bases__
             indexer = engine.Indexer(self.tempdir, 'w')
             for name, params in fixture.zipcodes.fields.items():
                 indexer.set(name, **params)
-            field = indexer.fields['tile'] = PointField('tile', precision=15, step=2, store=True)
+            with warnings.catch_warnings(record=True) as deprecations:
+                field = indexer.fields['tile'] = PointField('tile', precision=15, step=2, store=True)
+            assert deprecations or isinstance(field, engine.numeric.PointField)
             points = []
             for doc in fixture.zipcodes.docs():
                 if doc['state'] == 'CA':
@@ -381,13 +406,16 @@ class TestCase(BaseTest):
             hits = indexer.search(field.within(x, y, 10**5))
             cities = set(hit['city'] for hit in hits)
             assert city in cities and len(cities) > 100
+            assert len(field.within(x, y, 10**8)) == 1
             del indexer
     
     def testFields(self):
         "Custom field tests."
         indexer = engine.Indexer(self.tempdir)
         indexer.set('amendment', engine.FormatField, format='{0:02d}', store=True)
-        indexer.set('date', engine.DateTimeField, store=True)
+        with warnings.catch_warnings(record=True) as deprecations:
+            indexer.set('date', engine.DateTimeField, store=True)
+        assert deprecations
         indexer.set('size', engine.FormatField, format='{0:04d}', store=True)
         for doc in fixture.constitution.docs():
             if 'amendment' in doc:
@@ -404,8 +432,7 @@ class TestCase(BaseTest):
         assert str(query) == 'date:Y:{ TO 1921} date:Ym:[1921 TO 1921-12]'
         assert indexer.count(query) == 19
         query = field.range(datetime.date(1919, 1, 1), datetime.date(1921, 12, 31))
-        cls = lucene.TermRangeQuery if hasattr(lucene, 'TermRangeQuery') else lucene.ConstantScoreRangeQuery
-        fields = [cls.cast_(clause.query).field for clause in query]
+        fields = [lucene.TermRangeQuery.cast_(clause.query).field for clause in query]
         assert fields == ['date:Ymd', 'date:Ym', 'date:Y', 'date:Ym', 'date:Ymd']
         hits = indexer.search(query)
         assert [hit['amendment'] for hit in hits] == ['18', '19']
@@ -424,56 +451,73 @@ class TestCase(BaseTest):
         query = engine.Query.range('size', '1000', None)
         hits = indexer.search(query, sort=sizes.get)
         assert hits.ids == ids
-        hits = indexer.search(query, count=3, sort=lucene.SortField('size', lucene.SortField.LONG))
+        hits = indexer.search(query, count=3, sort=engine.SortField('size', type=long))
         assert hits.ids == ids[:len(hits)]
         query = engine.Query.range('size', None, '1000')
         assert indexer.count(query) == len(sizes) - len(ids)
+        indexer.sorters['year'] = engine.SortField('date:Ymd', type=int, parser=lambda date: int(date.split('-')[0]))
+        assert indexer.comparator('year')[:10] == [1791] * 10
+        hits = indexer.search(count=3, sort='year')
+        assert [int(hit['amendment']) for hit in hits] == [1, 2, 3]
+        cache = len(lucene.FieldCache.DEFAULT.cacheEntries)
+        hits = indexer.search(count=3, sort='year', reverse=True)
+        assert [int(hit['amendment']) for hit in hits] == [27, 26, 25]
+        assert cache == len(lucene.FieldCache.DEFAULT.cacheEntries)
+        indexer.add()
+        indexer.commit(sorters=True)
+        cache = len(lucene.FieldCache.DEFAULT.cacheEntries)
+        assert indexer.comparator('year')[-1] == 0
+        assert cache == len(lucene.FieldCache.DEFAULT.cacheEntries)
     
-    if numeric:
-        def testNumericFields(self):
-            "Numeric variant fields."
-            indexer = engine.Indexer(self.tempdir)
-            indexer.set('amendment', engine.numeric.NumericField, store=True)
-            indexer.set('date', engine.numeric.DateTimeField, store=True)
-            indexer.set('size', engine.numeric.NumericField, store=True)
-            for doc in fixture.constitution.docs():
-                if 'amendment' in doc:
-                    indexer.add(amendment=int(doc['amendment']), date=[map(int, doc['date'].split('-'))], size=len(doc['text']))
-            indexer.commit()
-            query = indexer.fields['amendment'].range(None, 10)
-            assert indexer.count(query) == 9
-            field = indexer.fields['date']
-            query = field.prefix((1791, 12))
-            assert indexer.count(query) == 10
-            query = field.prefix(datetime.date(1791, 12, 15))
-            assert indexer.count(query) == 10
-            query = field.range(None, (1921, 12), lower=False, upper=True)
-            assert indexer.count(query) == 19
-            query = field.range(datetime.date(1919, 1, 1), datetime.date(1921, 12, 31))
-            hits = indexer.search(query)
-            assert [hit['amendment'] for hit in hits] == ['18', '19']
-            assert [datetime.datetime.utcfromtimestamp(float(hit['date'])).year for hit in hits] == [1919, 1920]
-            query = field.within(seconds=100)
-            assert indexer.count(query) == 0
-            query = field.duration([2009], days=-100*365)
-            assert indexer.count(query) == 12
-            field = indexer.fields['size']
-            assert len(list(indexer.terms('size'))) > len(indexer)
-            sizes = dict((id, int(indexer[id]['size'])) for id in indexer)
-            ids = sorted((id for id in sizes if sizes[id] >= 1000), key=sizes.get)
-            query = field.range(1000, None)
-            hits = indexer.search(query, sort=sizes.get)
-            assert hits.ids == ids
-            hits = indexer.search(query, count=3, sort=lucene.SortField('size', lucene.SortField.LONG))
-            assert hits.ids == ids[:len(hits)]
-            query = field.range(None, 1000)
-            assert indexer.count(query) == len(sizes) - len(ids)
-            self.assertRaises(OverflowError, list, field.items(-2**64))
-            nf, = field.items(0.5)
-            assert nf.numericValue.doubleValue() == 0.5
-            assert str(field.range(-2**64, 0)) == 'size:[* TO 0}'
-            assert str(field.range(0, 2**64)) == 'size:[0 TO *}'
-            assert str(field.range(0.5, None, upper=True)) == 'size:[0.5 TO *]'
+    def testNumericFields(self):
+        "Numeric variant fields."
+        indexer = engine.Indexer(self.tempdir)
+        indexer.set('amendment', engine.numeric.NumericField, store=True)
+        indexer.set('date', engine.numeric.DateTimeField, store=True)
+        indexer.set('size', engine.numeric.NumericField, store=True, step=5)
+        for doc in fixture.constitution.docs():
+            if 'amendment' in doc:
+                indexer.add(amendment=int(doc['amendment']), date=[tuple(map(int, doc['date'].split('-')))], size=len(doc['text']))
+        indexer.commit()
+        query = indexer.fields['amendment'].range(None, 10)
+        assert indexer.count(query) == 9
+        field = indexer.fields['date']
+        query = field.prefix((1791, 12))
+        assert indexer.count(query) == 10
+        query = field.prefix(datetime.date(1791, 12, 15))
+        assert indexer.count(query) == 10
+        query = field.range(None, (1921, 12), lower=False, upper=True)
+        assert indexer.count(query) == 19
+        query = field.range(datetime.date(1919, 1, 1), datetime.date(1921, 12, 31))
+        hits = indexer.search(query)
+        assert [hit['amendment'] for hit in hits] == ['18', '19']
+        assert [datetime.datetime.utcfromtimestamp(float(hit['date'])).year for hit in hits] == [1919, 1920]
+        query = field.within(seconds=100)
+        assert indexer.count(query) == 0
+        query = field.duration([2009], days=-100*365)
+        assert indexer.count(query) == 12
+        field = indexer.fields['size']
+        assert len(list(indexer.terms('size'))) > len(indexer)
+        sizes = dict((id, int(indexer[id]['size'])) for id in indexer)
+        ids = sorted((id for id in sizes if sizes[id] >= 1000), key=sizes.get)
+        query = field.range(1000, None)
+        hits = indexer.search(query, sort=sizes.get)
+        assert hits.ids == ids
+        hits = indexer.search(query, count=3, sort=engine.SortField('size', type=long))
+        assert hits.ids == ids[:len(hits)]
+        query = field.range(None, 1000)
+        assert indexer.count(query) == len(sizes) - len(ids)
+        self.assertRaises(OverflowError, list, field.items(-2**64))
+        nf, = field.items(0.5)
+        assert nf.numericValue.doubleValue() == 0.5
+        assert str(field.range(-2**64, 0)) == 'size:[* TO 0}'
+        assert str(field.range(0, 2**64)) == 'size:[0 TO *}'
+        assert str(field.range(0.5, None, upper=True)) == 'size:[0.5 TO *]'
+        for step, count in zip(range(0, 20, field.step), (26, 19, 3, 1)):
+            sizes = list(indexer.numbers('size', step))
+            assert len(sizes) == count and all(isinstance(size, int) for size in sizes)
+            numbers = dict(indexer.numbers('size', step, type=float, counts=True))
+            assert sum(numbers.values()) == len(indexer) and all(isinstance(number, float) for number in numbers)
     
     def testHighlighting(self):
         "Highlighting text fragments."
